import os
import copy
import sys
from multiprocessing import Pool

import click

from midi_utils import chord
from saysynth import phoneme, say
from saysynth.cli.options import CHORD_OPTIONS, PHONEME_OPTIONS, SAY_OPTIONS, VELOCITY_OPTIONS, group_options, prepare_options_for_say


# # # # #
# Drone #
# # # # #


@click.command()
@click.argument("note", required=True)
@click.option(
    "-d",
    "--max-duration",
    default=1_000_000,
    type=int,
    help="The duration of the drone in milliseconds.",
)
@group_options(*CHORD_OPTIONS)
@group_options(*PHONEME_OPTIONS)
@click.option(
    "-pl",
    "--phoneme-loops",
    default=None,
    type=int,
    help="The number of times to loop an individual phoneme. If not provided, the drone will continue indefinitely.",
)
@group_options(*VELOCITY_OPTIONS)
@click.option(
    "-o",
    "--output-file",
    type=str,
    help="A filepath to write the generated text to",
)
@click.option('-x', '--exec', is_flag=True, default=False, help="Run the generated text through the say command. If `--chord` amd `--exec` are provided. ")
@group_options(*SAY_OPTIONS)
def run(**kwargs):
    """
    Given a note name (or midi note number), stream text required to generate a continuous drone for input to say
    """

    # generate the text if a chord isn't passed in
    if not kwargs.get('chord'):
        text = phoneme.drone_from_note(**kwargs)

        # handle writing text to file
        output_file = kwargs.get("output_file")
        if output_file:
            with open(output_file, "w") as f:
                f.write(text)

        # if we're not executing say, write text to stdout
        elif not kwargs.get('exec'):
            sys.stdout.write(text)

        else:
            # execute say
            kwargs = prepare_options_for_say(text, **kwargs)
            say.run(**kwargs)

    # spawn multiple commands, one per note in the chord
    else:
        kwargs['root'] = kwargs.pop('note')
        commands = []
        for note in chord.midi_chord(**kwargs):
            cmd_kwargs = copy.copy(kwargs)
            cmd_kwargs['text'] = phoneme.drone_from_note(note, type="note", **kwargs)
            commands.append(cmd_kwargs)
        say.spawn(commands)



import click

from midi_utils.constants import CHORDS
from saysynth.constants import (
    SAY_COLORS,
    SAY_EMPHASIS_VELOCITY_STEPS,
    SAY_ALL_PHONEMES,
    SAY_TUNED_VOICES,
    SAY_PHONEME_CLASSES,
    SAY_PHONEME_MAX_DURATION,
)


def group_options(*options):
    def wrapper(function):
        for option in reversed(options):
            function = option(function)
        return function

    return wrapper


def prepare_options_for_say(text: str, **kwargs):
    # handle some param edge cases
    rp = kwargs.get("randomize_phoneme")
    if rp and ":" in rp:
        kwargs["voice"] = rp.split(":")[0].strip().title()
    kwargs["text"] = text
    kwargs["use_tempfile"] = True
    return kwargs


# Phoneme Options

phoneme_opt = click.option(
    "-p",
    "--phoneme",
    default="m",
    help="The phoneme to use.",
    show_default=True,
    type=click.Choice(SAY_ALL_PHONEMES),
)
randomize_phoneme_opt = click.option(
    "-rp",
    "--randomize-phoneme",
    show_default=True,
    help=(
        "Randomize the phoneme for every note. "
        "If `all` is passed, all phonemes will be used. "
        "Alternatively pass a voice and style, eg: Fred:drone. "
        f"Valid voices include: {', '.join(SAY_TUNED_VOICES)}. "
        f"Valid styles include: {', '.join(SAY_PHONEME_CLASSES)}."
    ),
)

phoneme_duration_opt = click.option(
    "-pd",
    "--phoneme-duration",
    default=SAY_PHONEME_MAX_DURATION,
    show_default=True,
    type=int,
    help="The max duration in ms of an individual phoneme",
)

PHONEME_OPTIONS = [phoneme_opt, randomize_phoneme_opt, phoneme_duration_opt]

# Velocity Options

velocity_opt = click.option(
    "-vl",
    "--velocity",
    type=int,
    show_default=True,
    default=100,
    help="The midi velocity value to use for each note.",
)
velocity_steps_opt = click.option(
    "-vs",
    "--velocity-steps",
    type=int,
    nargs=2,
    show_default=True,
    default=SAY_EMPHASIS_VELOCITY_STEPS,
    help="The midi velocity values at which to add emphasis to a note",
)
randomize_velocity_opt = click.option(
    "-rv",
    "--randomize-velocity",
    type=int,
    nargs=2,
    help="Randomize a note's emphasis by supplying a min and max midi velocity (eg: -rv 40 120)",
)

VELOCITY_OPTIONS = [velocity_opt, velocity_steps_opt, randomize_velocity_opt]

# Say Options
rate_opt = click.option(
    "-r", "--rate", type=int, default=70, show_default=True, help="Rate to speak at"
)
voice_opt = click.option(
    "-v",
    "--voice",
    type=click.Choice(SAY_TUNED_VOICES),
    default="Fred",
    show_default=True,
    help="Voice to use",
)
# input_file_opt = click.option(
#     '-i',
#     "--input-file",
#     type=click.File(mode = "w"),
#     help="File to read text input from"
# )
output_file_opt = click.option(
    "-ao",
    "--audio-output-file",
    type=str,
    help="File to write audio output to",
)
audio_device_opt = click.option(
    "-ad",
    "--audio-device",
    type=str,
    help="Name of the audio device to send the signal to",
)
networks_send_opt = click.option(
    "-ns", "--network-send", type=str, help="Network address to send the signal to"
)
stereo_opt = click.option(
    "-st",
    "--stereo",
    is_flag=True,
    default=False,
    help="Whether or not to generate a stereo signal",
)
sample_size_opt = click.option(
    "-ss",
    "--sample-size",
    type=int,
    default=32,
    help="Sample size of the signal (1:32)",
)
sample_rate_opt = click.option(
    "-sr",
    "--sample-rate",
    type=int,
    default=22050,
    help="Sample rate of the signal (0:22050)",
)
quality_opt = click.option(
    "-q", "--quality", type=int, default=127, help="Quality of the signal (1:127)"
)
progress_bar_opt = click.option(
    "-pg",
    "--progress",
    is_flag=True,
    default=False,
    help="Whether or not to display an interactive progress bar",
)
interactive_opt = click.option(
    "-in",
    "--interactive",
    is_flag=True,
    default=False,
    help="Whether or not to display highlighted text",
)
text_color_opt = click.option(
    "-tc",
    "--text-color",
    type=click.Choice(SAY_COLORS),
    default="white",
    help="The text color to use when displaying highlighted text",
)
bg_color_opt = click.option(
    "-bc",
    "--bg-color",
    type=click.Choice(SAY_COLORS),
    default="black",
    help="The background color to use when displaying highlighted text",
)


SAY_OPTIONS = [
    rate_opt,
    voice_opt,
    output_file_opt,
    audio_device_opt,
    networks_send_opt,
    stereo_opt,
    sample_size_opt,
    sample_rate_opt,
    quality_opt,
    progress_bar_opt,
    interactive_opt,
    text_color_opt,
    bg_color_opt,
]

# Chord Options

chord_opt = click.option(
    "-c",
    "--chord",
    required=False,
    type=click.Choice([c.lower() for c in CHORDS.keys()]),
    help="An optional name of a chord to build using the note as root.",
)
inversion_opt = click.option(
    "-nv",
    "--inversion",
    default=[],
    required=False,
    type=lambda x: [int(i.strip()) for i in x.split(",")] if x else [],
    help="A list of integers (eg: '0,1,-1') specifying the direction and amplitude to invert each note. The length of this list much match the number of notes in the chord.",
)
stack_opt = click.option(
    "-k",
    "--stack",
    default=0,
    required=False,
    type=int,
    help="Stack a chord up (eg: '1' or '2') or down(eg: '-1' or '-2').",
)

CHORD_OPTIONS = [chord_opt, inversion_opt, stack_opt]



import os
import sys

import click

from saysynth import phoneme, say
from saysynth.cli.options import PHONEME_OPTIONS, group_options, randomize_velocity_opt, velocity_steps_opt, SAY_OPTIONS, prepare_options_for_say

# # # ##
# Midi #
# # # ##


@click.command()
@click.argument("midi_file", required=True)
@click.option(
    "-l",
    "--loops",
    default=1,
    show_default=True,
    type=int,
    help="The number of times to loop the midi file",
)
@click.option(
    "-oct",
    "--octave",
    type=int,
    help="The number of octaves to adjust the pitch up or down (eg: -oct 1 or -oct -1)",
)
@group_options(*PHONEME_OPTIONS)
@group_options(velocity_steps_opt, randomize_velocity_opt)
@click.option(
    "-o",
    "--output-file",
    type=str,
    help="A filepath to write the generated text to",
)
@click.option('-x', '--exec', is_flag=True, default=False, help="Run the generated text through the say command")
@group_options(*SAY_OPTIONS)
def run(**kwargs):
    """
    Given a midi file, generate a text-file (or stdout stream) of phonemes with pitch information for input to say
    """

    # generate the text
    text = phoneme.text_from_midi(**kwargs).replace("\n", " ")

    # handle writing text to file
    output_file = kwargs.get("output_file")
    if output_file:
        with open(output_file, "w") as f:
            f.write(text)
        sys.exit(0)

    # if we're not executing say, write text to stdout
    if not kwargs.get('exec'):
        sys.stdout.write(text)
        sys.exit(0)

    # execute say
    kwargs = prepare_options_for_say(text, **kwargs)
    say.run(**kwargs)



import random
from typing import Optional, List, Union, Tuple

from midi_utils import midi_to_freq, note_to_midi, freq_to_octave

from . import midi
from .utils import frange
from .constants import (
    SAY_PHONEME_SILENCE,
    SAY_PHONEME_MAX_DURATION,
    SAY_EMPHASIS_VELOCITY_STEPS,
    SAY_PHONEME_CLASSES,
    SAY_ALL_PHONEMES,
    SAY_TUNED_VOICES,
    SAY_TUNE_TAG,
    SAY_PHONEME_VOICE_CLASSES,
)


def get_emphasis(
    velocity: Optional[int],
    velocity_steps: Tuple[int, int] = SAY_EMPHASIS_VELOCITY_STEPS,
) -> str:
    """
    Translate a midi velocity value (0-127) into a phoneme emphasis value ("", "1", or "2")
    when provided with a tuple of steps (step_1, step_2) eg: (75, 100)
    """
    if not velocity:
        return ""
    if velocity > velocity_steps[1]:
        return "2"
    if velocity > velocity_steps[0]:
        return "1"
    return ""


def format_text(
    phoneme: str,
    duration: float,
    frequency: float,
    emphasis: int,
    is_silence: bool,
):
    """
    Format note text.
    TODO: Handle intra-note modulation? Use volm tag for velocity?
    """
    if is_silence:
        return f"{SAY_PHONEME_SILENCE} {{D {duration}}}"
    return f"{emphasis}{phoneme} {{D {duration}; P {frequency}:0 {frequency}:100}}"


def _get_phoneme_params(
    phoneme: str,
    randomize_phoneme: bool,
    velocity: int,
    randomize_velocity: bool,
    velocity_steps: Tuple[int, int],
) -> Tuple[str, int, int]:

    # handle phoneme randomization
    if randomize_phoneme:
        if randomize_phoneme == "all":
            phoneme = get_random()
        else:
            voice, style = randomize_phoneme.split(":")
            phoneme = get_random(voice, style)

    # handle velocity randomization
    if randomize_velocity:
        velocity = midi.random_velocity(*randomize_velocity)

    # generate emphasis
    emphasis = get_emphasis(velocity, velocity_steps)
    return phoneme, velocity, emphasis


def text_from_note(
    note: Union[int, str],
    duration: float,
    velocity: int,
    phoneme: str,
    type: str = "note",
    randomize_phoneme: Optional[str] = None,
    velocity_steps: Tuple[int, int] = SAY_EMPHASIS_VELOCITY_STEPS,
    randomize_velocity: Optional[Tuple[int, int]] = None,
    phoneme_duration: int = SAY_PHONEME_MAX_DURATION,
    octave: Optional[int] = None,
    sig_digits_duration: int = 2,
    randomize_segments: bool = False,
    **kwargs,
) -> str:
    """
    Generate say text for an individual note
    """
    is_silence = type == "silence"

    # lookup frequency
    if not is_silence:
        # allow note names
        frequency = midi_to_freq(note_to_midi(note))
        # octave adjustment
        if octave:
            frequency = freq_to_octave(frequency, octave)
    else:
        frequency = None

    phoneme, velocity, emphasis = _get_phoneme_params(
        phoneme, randomize_phoneme, velocity, randomize_velocity, velocity_steps
    )

    # return one note if the duration is short enough
    if duration <= phoneme_duration:
        return format_text(phoneme, duration, frequency, emphasis, is_silence)

    # otherwise create multiple phonemes which add up to the phoneme_duration
    texts = []
    for dur_step in frange(
        phoneme_duration, duration, phoneme_duration, sig_digits_duration
    ):

        # optionally randomize every segment.
        if randomize_segments:
            phoneme, velocity, emphasis = _get_phoneme_params(
                phoneme, randomize_phoneme, velocity, randomize_velocity, velocity_steps
            )

        text = format_text(phoneme, phoneme_duration, frequency, emphasis, is_silence)
        texts.append(text)

    # add final step and return
    text = format_text(phoneme, (duration - dur_step), frequency, emphasis, is_silence)
    texts.append(text)
    return f" ".join(texts) + " "


def drone_from_note(
    note: Union[int, str], velocity: int = 100, phoneme: str = "m", **kwargs
) -> str:
    """
    Generate drone text from `text_from_note` options
    """
    # set each phonemes duration to the max phoneme duration
    duration = kwargs["phoneme_duration"]
    max_duration = kwargs.get("max_duration")
    phoneme_loops = kwargs.get("phoneme_loops")
    loops = 0
    total_duration = 0
    drone_text = f"{SAY_TUNE_TAG} "
    while True:
        drone_text += f"{text_from_note(note, duration, velocity, phoneme, **kwargs)} "

        # check for max loops
        if phoneme_loops:
            loops += 1
            if loops >= phoneme_loops:
                break

        # check for max duration
        if max_duration:
            total_duration += duration
            if total_duration >= max_duration:
                break
    return drone_text


def text_from_midi(
    midi_file: str,
    phoneme: Optional[str] = None,
    randomize_phoneme: Optional[str] = None,
    phoneme_duration: int = SAY_PHONEME_MAX_DURATION,
    loops: int = 1,
    octave: Optional[int] = None,
    velocity_steps: Tuple[int, int] = SAY_EMPHASIS_VELOCITY_STEPS,
    randomize_velocity: Optional[Tuple[int, int]] = None,
    randomize_segments: bool = False,
    sig_digits_duration: int = 5,
    **kwargs,
) -> str:
    """
    Generate a text blob for input to say from a midi file, optionally looping it.
    TODO: vary phonemes
    """
    text = f"{SAY_TUNE_TAG} "
    for _ in range(0, loops, 1):
        for note in midi.process(midi_file):
            text += text_from_note(
                phoneme=phoneme,
                phoneme_duration=phoneme_duration,
                octave=octave,
                velocity_steps=velocity_steps,
                randomize_velocity=randomize_velocity,
                randomize_phoneme=randomize_phoneme,
                sig_digits_duration=sig_digits_duration,
                randomize_segments=randomize_segments,
                **note,
            )
    return text


def get_random(
    voice: Optional[str] = None,
    style: Optional[str] = None,
    choices: Optional[List[str]] = None,
) -> str:
    """
    Generate a random phoneme.
    Args:
        voice: Either Fred, Alex, or Veronica
        style: Either drone, noise, or note
        choices: A list of phonemes to choose from
    """
    if choices:
        return random.choice(choices)
    if (style and not voice) or (voice and not style):
        raise ValueError(
            "To use `style`/`voice` randomization, both parameters must be supplied."
        )
    if style and voice:
        voice = voice.title()  # allow for lowercase
        try:
            return random.choice(SAY_PHONEME_VOICE_CLASSES[voice][style])
        except KeyError:
            raise ValueError(
                f"Invalid `voice` '{voice}' or `style` '{style}'. "
                f"`voice` must be one of: {', '.join(SAY_TUNED_VOICES)}. "
                f"`style` must be one of: {', '.join(SAY_PHONEME_CLASSES)}"
            )
    return random.choice(SAY_ALL_PHONEMES)



import os
import math
import tempfile

# midi / note / freq lookup


def frange(start, stop, by, sig_digits=5):
    """
    Generate a range of float values.
    """
    div = math.pow(10, sig_digits)
    for value in range(int(start * div), int(stop * div), int(by * div)):
        yield round(value / div, sig_digits)


def here(f, *args):
    """
    Pass __file__ to get the current directory and *args to generate a filepath
    """
    return os.path.join(os.path.dirname(os.path.abspath(f)), *args)


def make_tempfile(format: str = "txt"):
    return tempfile.mkstemp(suffix=f".{format}")[-1]




import os

import click
from midi_utils import midi_to_note, note_to_midi, midi_scale
from midi_utils import ROOT_TO_MIDI, SCALES

from saysynth import say, phoneme, utils
from saysynth.constants import (
    SAY_TUNE_TAG,
)
from saysynth.cli.options import group_options, VELOCITY_OPTIONS, PHONEME_OPTIONS, voice_opt, rate_opt

@click.command()
@click.option(
    "-s", "--start-at", type=str, default="C3", help="Note name/number to start at"
)
@click.option(
    "-e",
    "--end-at",
    type=str,
    default="G5",
    show_default=True,
    help="Note name/number to end at",
)
@click.option(
    "-c",
    "--scale",
    type=click.Choice([s.lower() for s in SCALES]),
    default="minor",
    show_default=True,
    help="Scale name to use",
)
@click.option(
    "-k", "--key", type=click.Choice(ROOT_TO_MIDI.keys()), default="C", show_default=True, help="Root note of scale"
)
@group_options(voice_opt, rate_opt)
@click.option(
    "-o",
    "--output-dir",
    default="./",
    type=str,
    show_default=True,
    help="Directory to write to",
)
@click.option(
    "-f",
    "--format",
    type=click.Choice(["wav", "aiff"]),
    default="aiff",
    show_default=True,
    help="Format of each note's file.",
)
@group_options(*PHONEME_OPTIONS)
@click.option(
    "-d",
    "--duration",
    default=1_000_000,
    type=int,
    help="The duration each note of the soundfont in milliseconds.",
)
@group_options(*VELOCITY_OPTIONS)
@click.option(
    "-rs",
    "--randomize-segments",
    type=bool,
    default=False,
    help="Randomize every segment of a note. Only applies when --randomize-velocity and --randomize-phoneme are set.",
)
def run(**kwargs):
    """
    Given a scale and other parameters, generate a soundfont of each note as an .aiff or .wav file.
    """
    # add note type to simplify function call
    kwargs["type"] = "note"

    # determine set of notes to generate
    start_at = note_to_midi(kwargs["start_at"])
    end_at = note_to_midi(kwargs["end_at"])
    scale = midi_scale(
        key=kwargs["key"], scale=kwargs["scale"], min_note=start_at, max_note=end_at
    )

    # generate files for each note in the scale
    for midi in scale:

        # add note type/midi not to simplify phoneme_text_from_note function call
        kwargs["type"] = "note"
        kwargs["midi"] = midi
        kwargs["note"] = midi_to_note(midi)

        # generate audio output file name
        audio_output_file = os.path.join(
            kwargs["output_dir"],
            f"{kwargs['voice'].lower()}-{kwargs['rate']}-{midi:02d}.{kwargs['format']}",
        )
        # generate input file of text
        input_file_name = utils.make_tempfile()
        with open(input_file_name, "w") as f:
            f.write(f"{SAY_TUNE_TAG}\n{phoneme.text_from_note(**kwargs)}")
        cmd = say.cmd(
            input_file=input_file_name,
            voice=kwargs["voice"],
            rate=kwargs["rate"],
            audio_output_file=audio_output_file,
        )
        say.run(cmd)
        if not os.path.exists(audio_output_file):
            raise RuntimeError(f"File {audio_output_file} was not successfully created")
        # cleansup tempfile
        os.remove(input_file_name)



import os
from multiprocessing import Pool
import subprocess
from this import d
from typing import Optional, List, Dict, Any

from . import utils
from .constants import (
    SAY_MAX_SAMPLE_RATE,
    SAY_SAMPLE_SIZES,
    SAY_DATA_TYPES,
    SAY_ENDIANESS,
    SAY_EXECUTABLE,
    SAY_COLORS,
)


def _gen_data_format_arg(
    endianess: str, data_type: str, sample_size: int, sample_rate: int
):
    """
    Generate a string to pass to --data-format
    """
    if endianess not in SAY_ENDIANESS:
        raise ValueError(
            "Invalid `endianess`. Choose from: LE (little endlian) or BE (big endian)"
        )
    if data_type not in SAY_DATA_TYPES:
        raise ValueError(
            "Invalid `data_type`. Choose from: F (float), I (integer), UI (unsigned integer)"
        )
    if sample_size not in SAY_SAMPLE_SIZES:
        raise ValueError(
            f'Invalid `sample_size`. Choose from: {", ".join(SAY_SAMPLE_SIZES)}'
        )

    # allow pass passing sample rate as small number (eg: 24 -> 24000)
    if sample_rate < 1000:
        sample_rate *= 1000

    # don't allow a sample rate greater than the maximum
    if sample_rate > SAY_MAX_SAMPLE_RATE:
        sample_rate = SAY_MAX_SAMPLE_RATE

    return f"{endianess}{data_type}{sample_size}@{int(sample_rate)}"


def _gen_interactive_arg(text_color: str = "white", bg_color: str = "black"):
    """
    Generate a string to pass to --interactive
    """
    if bg_color and not text_color:
        text_color = "white"  # default text color if only background is supplied
    if text_color not in SAY_COLORS:
        raise ValueError(f'Invalid `text_color`, choose from: {", ".join(SAY_COLORS)}')
    if bg_color not in SAY_COLORS:
        raise ValueError(f'Invalid `bg_color`, choose from: {", ".join(SAY_COLORS)}')
    return f"--interactive={text_color}/{bg_color}"


def cmd(
    text: Optional[str] = None,
    voice: Optional[str] = None,
    rate: Optional[int] = None,
    input_file: Optional[str] = None,
    audio_output_file: Optional[str] = None,
    service_name: Optional[str] = None,
    network_send: Optional[str] = None,
    audio_device: Optional[str] = None,
    stereo: bool = False,  # whether or not
    endianness: str = "LE",  # LE/BE
    data_type: str = "F",  # F/I/U
    sample_size: Optional[int] = 32,
    sample_rate: Optional[int] = 22050,
    quality: int = 127,
    progress: bool = False,
    interactive: bool = False,
    text_color: Optional[str] = None,
    bg_color: Optional[str] = None,
    executable: str = SAY_EXECUTABLE,
    **kwargs,
) -> List:
    """
    A python wrapper around the say command.
    OPTIONS
        string
            Specify the text to speak on the command line. This can consist of multiple arguments, which are considered to be separated
            by spaces.

        -f file, --input-file=file
            Specify a file to be spoken. If file is - or neither this parameter nor a message is specified, read from standard input.

        -v voice, --voice=voice
            Specify the voice to be used. Default is the voice selected in System Preferences. To obtain a list of voices installed in
            the system, specify '?' as the voice name.

        -r rate, --rate=rate
            Speech rate to be used, in words per minute.

        -o out.aiff, --output-file=file
            Specify the path for an audio file to be written. AIFF is the default and should be supported for most voices, but some
            voices support many more file formats.

        -n name, --network-send=name
        -n name:port, --network-send=name:port
        -n :port, --network-send=:port
        -n :, --network-send=:
            Specify a service name (default "AUNetSend") and/or IP port to be used for redirecting the speech output through AUNetSend.

        -a ID, --audio-device=ID
        -a name, --audio-device=name
            Specify, by ID or name prefix, an audio device to be used to play the audio. To obtain a list of audio output devices,
            specify '?' as the device name.

        --progress
            Display a progress meter during synthesis.

        -i, --interactive, --interactive=markup
            Print the text line by line during synthesis, highlighting words as they are spoken. Markup can be one of

            o   A terminfo capability as described in terminfo(5), e.g. bold, smul, setaf 1.

            o   A color name, one of black, red, green, yellow, blue, magenta, cyan, or white.

            o   A foreground and background color from the above list, separated by a slash, e.g. green/black. If the foreground color
                is omitted, only the background color is set.

            If markup is not specified, it defaults to smso, i.e. reverse video.

        If the input is a TTY, text is spoken line by line, and the output file, if specified, will only contain audio for the last line
        of the input.  Otherwise, text is spoken all at once.

    AUDIO FORMATS
        Starting in MacOS X 10.6, file formats other than AIFF may be specified, although not all third party synthesizers may initially support them. In simple cases, the
        file format can be inferred from the extension, although generally some of the options below are required for finer grained control:

        --file-format=format
            The format of the file to write (AIFF, caff, m4af, WAVE). Generally, it's easier to specify a suitable file extension for the output file. To obtain a list of
            writable file formats, specify '?' as the format name.

        --data-format=format
            The format of the audio data to be stored. Formats other than linear PCM are specified by giving their format identifiers (aac, alac). Linear PCM formats are
            specified as a sequence of:

            Endianness (optional)
                One of BE (big endian) or LE (little endian). Default is native endianness.

            Data type
                One of F (float), I (integer), or, rarely, UI (unsigned integer).

            Sample size
                One of 8, 16, 24, 32, 64.

            Most available file formats only support a subset of these sample formats.

            To obtain a list of audio data formats for a file format specified explicitly or by file name, specify '?' as the format name.

            The format identifier optionally can be followed by @samplerate and /hexflags for the format.

        --channels=channels
            The number of channels. This will generally be of limited use, as most speech synthesizers produce mono audio only.

        --bit-rate=rate
            The bit rate for formats like AAC. To obtain a list of valid bit rates, specify '?' as the rate. In practice, not all of these bit rates will be available for a
            given format.

        --quality=quality
            The audio converter quality level between 0 (lowest) and 127 (highest).

    """
    if not text and not input_file:
        raise ValueError("Must provide `text` or `input_file`")

    # override text if input file is provided
    elif input_file:
        text = None
        # verify that input file exists
        if not os.path.exists(input_file):
            raise ValueError("`input_file`: {input_file} does not exist!")

    # verify file_format:
    if audio_output_file:
        if not (
            audio_output_file.lower().endswith(".wav")
            or audio_output_file.lower().endswith(".aiff")
        ):
            raise ValueError("`output_file` must be of type .wav or .aiff")

    # verify quality
    if quality < 0 or quality > 127:
        raise ValueError("`quality` must be between 0 and 127")

    # construct base command
    cmd = [executable]
    if text:
        cmd.append(text)
    elif input_file:
        cmd.extend(["-f", input_file])
    if voice:
        cmd.extend(["-v", voice])
    if rate:
        cmd.extend(["-r", rate])

    # output_file specific params
    if audio_output_file:
        cmd.extend(["-o", audio_output_file])
        # # always specify data format
        # cmd.append(
        #     f"--data-format={_gen_data_format_arg(endianness, data_type, sample_size, sample_rate)}"
        # )
        if stereo:
            cmd.append(f"--channels=2")
    else:
        cmd.append(f"--quality={quality}")
        # handle network output if output file is not specified
        if service_name:
            cmd.extend(["-n", service_name])
        if network_send:
            cmd.extend(f"--network-send={network_send}")
        if audio_device:
            cmd.extend(["-a", audio_device])

    # progress bar
    if progress:
        cmd.append("--progress")

    # interactivity
    if interactive:
        cmd.append(_gen_interactive_arg(text_color, bg_color))
    return [str(a) for a in cmd]


def run(args: Optional[List] = None, use_tempfile: bool = True, **kwargs):
    """
    Execute a command given a list of arguments () outputted by `cmd`
    or by supplying the kwargs that `cmd` accepts
    """
    # write text as input file
    tempfile = None
    if kwargs.get("text") and use_tempfile:
        text = kwargs.pop("text")
        tempfile = utils.make_tempfile()
        with open(tempfile, "w") as f:
            f.write(text)
        kwargs["input_file"] = tempfile

    if not args:
        args = cmd(**kwargs)
    try:
        p = subprocess.Popen(args, stdout=subprocess.PIPE)
        p.wait()
    finally:
        if tempfile:
            os.remove(tempfile)


def _run_fx(kwargs: Dict[str, Any]):
    """
    utility for passing kwargs into run fx
    """
    return run(**kwargs)


def spawn(commands: List[Dict[str, Any]]):
    """
    Spawn multiple say processes in parallel
    """
    p = Pool(len(commands))
    p.map(_run_fx, commands)



from setuptools import setup, find_packages

config = {
    "name": "saysynth",
    "version": "0.0.7",
    "packages": find_packages(),
    "install_requires": ["charset-normalizer", "click", "mido", "midi-utils"],
    "author": "Brian Abelson",
    "author_email": "hey@gltd.email",
    "description": "A musical generator for say",
    "url": "http://globally.ltd",
    "entry_points": {
        "console_scripts": ["saysynth=saysynth.cli:main", "sy=saysynth.cli:main"]
    },
}

setup(**config)



# # # # # # # # # #
# Say Constants
# # # # # # # # # #

# the path to where the say command exists
SAY_EXECUTABLE = "/usr/bin/say"

SAY_FILE_FORMATS = ["wav", "aiff"]

# voices which respect [[TUNE]] input
SAY_TUNED_VOICES = ["Alex", "Fred", "Victoria"]

# opening tag for tuned input to say
SAY_TUNE_TAG = "[[inpt TUNE]]"

# colors which can used to style the interactive output
SAY_COLORS = ["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"]

# values for --data-format to determine the endianess
SAY_ENDIANESS = ["BE", "LE"]

# values for --data-format to determine the data type
SAY_DATA_TYPES = ["F", "I", "UI"]

# values for --data-format to determine the sample size
SAY_SAMPLE_SIZES = [8, 16, 24, 32, 64]

# this is the max sample rate, anything above this will
# generate upsampled audio
# from: https://stackoverflow.com/questions/9729153/error-on-say-when-output-format-is-wave
SAY_MAX_SAMPLE_RATE = 22050

# # # # # # # # # #
# Phoneme Constants
# # # # # # # # # #
# Taken from: https://developer.apple.com/library/archive/documentation/UserExperience/Conceptual/SpeechSynthesisProgrammingGuide/Phonemes/Phonemes.html#//apple_ref/doc/uid/TP40004365-CH9-SW1

# these phonemes can create continuous pitches
SAY_ALL_PHONEMES = [
    "AE",
    "EY",
    "AO",
    "AX",
    "IY",
    "EH",
    "IH",
    "AY",
    "IX",
    "AA",
    "UW",
    "UH",
    "UX",
    "OW",
    "AW",
    "OY",
    "b",
    "C",
    "d",
    "D",
    "f",
    "g",
    "h",
    "J",
    "k",
    "l",
    "m",
    "n",
    "N",
    "p",
    "r",
    "s",
    "S",
    "t",
    "T",
    "v",
    "w",
    "y",
    "z",
    "Z",
]

# generated by running scripts/classify_phonemes.py
SAY_PHONEME_CLASSES = ["drone", "noise", "note"]
SAY_PHONEME_VOICE_CLASSES = {
    "Alex": {
        "drone": [
            "AE",
            "EY",
            "AO",
            "AX",
            "IY",
            "EH",
            "IH",
            "AY",
            "IX",
            "UW",
            "OW",
            "OY",
            "h",
            "l",
            "m",
            "n",
            "N",
            "r",
        ],
        "noise": ["C", "d", "D", "f", "J", "k", "p", "s", "S", "t", "T", "v", "z"],
        "note": ["AA", "UH", "UX", "AW", "b", "g", "w", "y", "Z"],
    },
    "Fred": {
        "drone": [
            "AE",
            "EY",
            "AO",
            "AX",
            "IY",
            "EH",
            "IH",
            "AY",
            "IX",
            "AA",
            "UW",
            "UH",
            "UX",
            "OW",
            "AW",
            "OY",
            "D",
            "l",
            "m",
            "n",
            "N",
            "r",
            "v",
            "w",
            "y",
            "z",
            "Z",
        ],
        "note": [  # all of Fred's drones work as notes.
            "AE",
            "EY",
            "AO",
            "AX",
            "IY",
            "EH",
            "IH",
            "AY",
            "IX",
            "AA",
            "UW",
            "UH",
            "UX",
            "OW",
            "AW",
            "OY",
            "D",
            "l",
            "m",
            "n",
            "N",
            "r",
            "v",
            "w",
            "y",
            "z",
            "Z",
        ],
        "noise": ["b", "C", "d", "f", "g", "h", "J", "k", "p", "s", "S", "t", "T"],
    },
    "Victoria": {
        "drone": [
            "AE",
            "EY",
            "AO",
            "AX",
            "IY",
            "EH",
            "IH",
            "AY",
            "IX",
            "AA",
            "UW",
            "UH",
            "UX",
            "AW",
            "OY",
            "l",
            "m",
            "n",
            "N",
            "r",
            "v",
            "w",
            "y",
        ],
        "noise": ["C", "d", "D", "f", "h", "k", "p", "s", "S", "t", "T", "z", "Z"],
        "note": ["OW", "b", "g", "J"],
    },
}

SAY_PHONEME_SILENCE = "%"

# this is about the number of
# milliseconds for an individual phoneme
# at which the duration of output stops changing
SAY_PHONEME_MAX_DURATION = 1200

# the midi velocity values above which
# we add an emphasis to a phoneme
SAY_EMPHASIS_VELOCITY_STEPS = [75, 100]
